/*
Module : ATLCPImplMT.h
Purpose: Implementation for a replacement class for the ATL class IConnectionPointImpl which performs 
         automatic marshaling of client interfaces acrosss COM apartments using the GIT (Global Interface
         Table). Put more simply, this class will allow you to fire events from your COM objects from
         a different thread than the one on which the object was initially created. The code was based
         on the MS Article "SAMPLE: ATLCPImplMT encapsulates ATL event firing across COM apartments" at
         http://support.microsoft.com/kb/280512. As mentioned in this article, because the code uses the 
         Global Interface Table, it requires Windows NT 4.0 Service Pack 3 or later, or Windows 95 (or later) 
         with DCOM version 1.1 or later to be installed. Based on this original code, I have performed an 
         extensive rewrite of the class using the ATL 8 "CComGITPtr" class. In addition it has been updated 
         based upon the ATL 8 "IConnectionPointImpl" implementation. It also uses ATL::CSimpleArray
         to hold the GIT'ed interfaces rather than the CComDynamicUnkArray class.
         
         Note that this class unlike the rest of the code in GPSCom is licensed as open-source / freeware, 
         meaning that you are free to use this class in your projects (subject to the copyright details 
         below), withour being required to register GPSCom.
         
         To use the class in your ATL client code, you first need to perform the usual sequence of steps
         to support events in your ATL project. This will result in the creation of a proxy class which is
         derived from IConnectionPointImpl which contains a number of "Fire_{EventName}" methods which you 
         would then derive from in your normal ATL objects, thus allowing you to fire events back to your 
         clients. You should change the inheritance of this proxy class to be derived from IConnectionPointImplMT 
         instead of IConnectionPointImpl. Note that the template parameters for IConnectionPointImplMT are 
         slightly different than IConnectionPointImpl. The template parameters for IConnectionPointImplMT 
         are as follows:
         
         template <class T, const IID* piid, class TSinkInterface> class IConnectionPointImplMT.
         
         Also for each generated function (named Fire_{EventName}) within the proxy class, you will need to          
         rework the firing code similiar to the following: 

         HRESULT Fire_OnRMCSentence()
	       {
		       CComVariant varResult;
		       int nConnections = m_Clients.GetSize();
		       for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
		       {
             CComPtr<IDispatch> sp(GetInterfaceAt(nConnectionIndex));
			       if (sp)
			       {
			         Code to call sp->Invoke with correct parameters....
			         ......
			         ......
				       sp->Invoke(.....);
			        }
		        }
		        return .....;
	        }
         
	       
	       That's all you need to do. All other client remains the same. For a concreate example of using the 
	       class in client code, please see the GPSComCP.h file in the GPSCom download. Happy COM coding!
	       
Created: PJN / 21-06-2002
History: PJN / 20-07-2006 1. Updated code to clean compile on VC 2005
         PJN / 20-01-2007 1. Major rewrite of the code as documented above in the Purpose section

Copyright (c) 2002 - 2007 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#pragma once

#ifndef __CPIMPLMT_H__
#define __CPIMPLMT_H__


////////////////////////// Includes ///////////////////////////////////////////

#include <atlcom.h>


///////////////////////// Classes /////////////////////////////////////////////

template <class T, const IID* piid, class TSinkInterface>
class ATL_NO_VTABLE IConnectionPointImplMT : public _ICPLocator<piid>
{
	typedef CComEnum<IEnumConnections, &__uuidof(IEnumConnections), CONNECTDATA, 
	  _Copy<CONNECTDATA> > CComEnumConnections;

public:
//Methods
	STDMETHOD(_LocCPQueryInterface)(REFIID riid, void** ppvObject)
	{
#ifndef _ATL_OLEDB_CONFORMANCE_TESTS
		ATLASSERT(ppvObject != NULL);
#endif
		if (ppvObject == NULL)
			return E_POINTER;
		*ppvObject = NULL;

		if (InlineIsEqualGUID(riid, __uuidof(IConnectionPoint)) || InlineIsEqualUnknown(riid))
		{
			*ppvObject = this;
			AddRef();
#ifdef _ATL_DEBUG_INTERFACES
			_AtlDebugInterfacesModule.AddThunk(reinterpret_cast<IUnknown**>(ppvObject), _T("IConnectionPointImplMT"), riid);
#endif // _ATL_DEBUG_INTERFACES
			return S_OK;
		}
		else
			return E_NOINTERFACE;
	}

	STDMETHOD(GetConnectionInterface)(IID* piid2)
	{
		if (piid2 == NULL)
			return E_POINTER;
		*piid2 = *piid;
		return S_OK;
	}

	STDMETHOD(GetConnectionPointContainer)(IConnectionPointContainer** ppCPC)
	{
		T* pT = static_cast<T*>(this);
		
		//No need to check ppCPC for NULL since QI will do that for us
		return pT->QueryInterface(IID_IConnectionPointContainer, reinterpret_cast<void**>(ppCPC));
	}

	STDMETHOD(Advise)(IUnknown* pUnkSink, DWORD* pdwCookie)
  {
	  //Validate our parameter
	  if (pUnkSink == NULL || pdwCookie == NULL)
		  return E_POINTER;
		
		//Default the output parameters
		*pdwCookie = 0;

    //Check that the client implements the expected sink interface
    CComPtr<TSinkInterface> pClientSink;
    HRESULT hr = pUnkSink->QueryInterface(&pClientSink);  
	  if (SUCCEEDED(hr))
	  {
	    //Marshal the interface into the GIT
		  CComGITPtr<TSinkInterface> newClient;
		  hr = newClient.Attach(pClientSink);
		  if (hr == S_OK)
		  {
		    //And then get the corresponding cookie for the specified interface
		    *pdwCookie = newClient.GetCookie();
		    
		    //Serialize access to the array as we are about to add a new client
  		  CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);
  		  
  		  //Finally add to the array of clients
        m_Clients.Add(newClient);        
		  }
	  }
	  else if (hr == E_NOINTERFACE)
		  hr = CONNECT_E_CANNOTCONNECT;

	  return hr;
  }

	STDMETHOD(Unadvise)(DWORD dwCookie)
  {
    //Serialize access to the array
    CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);

    //What will be the return value from this function (assume the best)
    HRESULT hr = S_OK;
    
    //Remove the specified client from the array of clients, by matching up
    //the input cookie with the cookies in our clients array
    BOOL bFound = FALSE;
    int nClients = m_Clients.GetSize();
    for (int i=0; i<nClients && !bFound; i++)
    {
      CComGITPtr<TSinkInterface>& client = m_Clients[i];
      if (client.GetCookie() == dwCookie)
      {
        bFound = TRUE;
        hr = client.Revoke();
        m_Clients.RemoveAt(i);
      }
    }

	  return SUCCEEDED(hr) ? S_OK : CONNECT_E_NOCONNECTION;
  }

	STDMETHOD(EnumConnections)(IEnumConnections** ppEnum)
  {
    //Validate our parameters
	  if (ppEnum == NULL)
		  return E_POINTER;

    //Set the output parameter to a default value
	  *ppEnum = NULL;

    //Create the COM object which will handle the COM interface being returned
	  CComObject<CComEnumConnections>* pEnum = NULL;
	  ATLTRY(pEnum = new CComObject<CComEnumConnections>)
	  if (pEnum == NULL)
		  return E_OUTOFMEMORY;

    //Serialize access to the clients array
	  CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);
	  
	  //Allocate the array
	  CONNECTDATA* pcd = NULL;
	  int nClients = m_Clients.GetSize();
	  ATLTRY(pcd = new CONNECTDATA[nClients])
	  if (pcd == NULL)
	  {
		  delete pEnum;
		  return E_OUTOFMEMORY;
	  }

	  //Copy the valid CONNECTDATA's
	  CONNECTDATA* pend = pcd;
	  HRESULT hr = S_OK;
	  for (int i=0; i<nClients; i++)
	  {
		  CComGITPtr<TSinkInterface>& client = m_Clients[i];
		  CComPtr<TSinkInterface> sinkInterface;
	    hr = client.CopyTo(&sinkInterface);
	    if (SUCCEEDED(hr))
	    {
	      hr = sinkInterface.QueryInterface(&pend->pUnk);
	      if (SUCCEEDED(hr))
	      {
	        //Also set the cookie value and move on to the next element in the output CONNECTDATA array
		      pend->dwCookie = client.GetCookie();
		      pend++;
		    }
	    }
	  }
	  //don't copy the data, but transfer ownership to it
	  hr = pEnum->Init(pcd, pend, NULL, AtlFlagTakeOwnership);
	  lock.Unlock();
	  if (FAILED(hr))
	    delete pEnum;
	  else
	  {
	    hr = pEnum->_InternalQueryInterface(IID_IEnumConnections, reinterpret_cast<void**>(ppEnum));
	    if (FAILED(hr))
		    delete pEnum;
		}
	  return hr;
  }

	IDispatch* GetInterfaceAt(int nConnectionIndex)
  {
    //Serialize access to the client array
	  CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);

	  CComPtr<TSinkInterface> sinkInterface;
	  HRESULT hr = m_Clients[nConnectionIndex].CopyTo(&sinkInterface);
	  
	  ATLASSERT(SUCCEEDED(hr));
	  IDispatch* pDisp = NULL;
	  hr = sinkInterface.QueryInterface(&pDisp);
	  ATLASSERT(SUCCEEDED(hr));

	  return pDisp;
  }

//Member variables
	CSimpleArray<CComGITPtr<TSinkInterface > > m_Clients;      //The array of GIT interface pointers we are managing
	CComGlobalsThreadModel::AutoCriticalSection m_CPMTCritSec; //The critical section to serialize access to "m_Clients" above
};

#endif //__CPIMPLMT_H__
