//////////////////// Includes ///////////////////////////////////

#include "stdafx.h"
#include "RedRat.h"
#include "RedRatConnection.h"


//////////////////// Macros / Defines ///////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

const long AUTO_RECEIVE_EEPROM_LOCATION = 4;


//////////////////// Implementation /////////////////////////////

CRedRatConnection::CRedRatConnection() : m_pMonitorThread(NULL),
                                         m_nPort(0),
                                         m_lTimeout(5000),
                                         m_bAutoReceiveSignals(FALSE)
{
  memset(&m_Overlapped, 0, sizeof(OVERLAPPED));
  m_Overlapped.hEvent = m_PortEvent;
}

CRedRatConnection::~CRedRatConnection()
{
  //Ensure we close down everything before we are finished
  Close();

  //Synchronize access to the connections array
  CMyApp* pApp = GetApp();
  CComCritSecLock<CComAutoCriticalSection> lock(pApp->m_csConnections);

  //Remove this instance from the Module list
  BOOL bFound = FALSE;
  int nConnections = pApp->m_Connections.GetSize();
  for (int i=0; i<nConnections && !bFound; i++)
  {
    CComObject<CRedRatConnection>* pConnection = static_cast<CComObject<CRedRatConnection>*>(pApp->m_Connections[i]);
    if (pConnection->m_nPort == m_nPort)
    {
      pApp->m_Connections.RemoveAt(i);
      bFound = TRUE;
    }
  }
}

HRESULT CRedRatConnection::Open(long nPort)
{
  //Open up the serial port
  try
  {
    //Validate out state
    ATLASSERT(!m_Port.IsOpen());

    //Open up the serial port using the standard settings for the RedRat
    m_Port.Open(nPort, 19200, CSerialPort::NoParity, 8, CSerialPort::OneStopBit, CSerialPort::NoFlowControl, TRUE);

    //We are interested in data being received
    m_Port.SetMask(EV_RXCHAR);

    m_Port.Set0ReadTimeout();
  }
  catch(CSerialException* pEx)
  {
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    m_Port.Close();

    //Form the rich error information
    return Error(IDS_FAIL_OPEN_PORT, IID_IRedRatConnection, hr);
  }

  //Store away the member variables
  m_nPort = nPort;

  return S_OK;
}

void CRedRatConnection::Close()
{
  TerminateMonitoringThread();

  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  //Close down the comms port
  m_Port.Close();
}

UINT CRedRatConnection::_MonitorThread(void* pVoid)
{
  //Validate our parameters
  ATLASSERT(pVoid);

  //Initialize COM in this thread as we will be using connection points from it
  CoInitialize(NULL);

  //Convert from the SDK world to the C++ world
  CRedRatConnection* pThis = static_cast<CRedRatConnection*>(pVoid);
  ATLASSERT(pThis);
  UINT nReturn = pThis->MonitorThread();

  //Unitialize COM in this thread
  CoUninitialize();

  return nReturn;
}

BOOL CRedRatConnection::ReadResponse(CString& sResponse, BOOL& bError, CString& sErrorDescription)
{
  sResponse.Empty();
  sErrorDescription.Empty();
  bError = FALSE;

  BOOL bSuccess = FALSE;
  BOOL bMore = TRUE;

  while (bMore)
  {
    try
    {
      DWORD dwEvents = 0;
      m_Port.WaitEvent(dwEvents, m_Overlapped);
    }
    catch(CSerialException* pEx)
    {
      pEx->Delete();
    }

    DWORD dwWait = WaitForSingleObject(m_PortEvent, m_lTimeout);
    if (dwWait == WAIT_OBJECT_0)
    {
      char szBuffer[100];
      DWORD dwBytesRead = 0;
      try
      {
        if (!m_Port.Read(szBuffer, sizeof(szBuffer) - 1, m_Overlapped, &dwBytesRead))
          m_Port.GetOverlappedResult(m_Overlapped, dwBytesRead, TRUE);
      }
      catch(CSerialException* pEx)
      {
        pEx->Delete();
      }

      //If any data was read from the serial port, then parse the received data
      if (dwBytesRead > 0) 
      {
        //Null terminate the received data
        szBuffer[dwBytesRead] = '\0';

        //Append the newly read data to the buffer
        sResponse += szBuffer;

        //See if we've got a full request yet
        int nLocationResponseTerminator = sResponse.Find(_T("\r\n>"));
        bMore = (nLocationResponseTerminator == -1);
      }
    }
    else
      bMore = FALSE;
  }

  //Tidy up the message
  int nMessageLength = sResponse.GetLength();
  if (nMessageLength)
  {
    //Remove the start message indicator
    int nMessageStart = sResponse.Find(_T('['));
    if (nMessageStart == 0)
    {
      bSuccess = TRUE;
      nMessageLength--;
      sResponse = sResponse.Right(nMessageLength);
    }

    //Remove the end message indicator
    int nMessageEnd = sResponse.ReverseFind(_T(']'));
    if (nMessageEnd != -1)
    {
      nMessageLength = nMessageEnd;
      sResponse = sResponse.Left(nMessageEnd);
    }

    //Remove the optional checksum
    int nChecksum = sResponse.ReverseFind(_T('#'));
    if (nChecksum != -1)
      sResponse = sResponse.Left(nChecksum);

    //Is it an error code
    int nExclaimationPos = sResponse.Find(_T('!'));
    if (nExclaimationPos != -1)
    {
      bError = TRUE;

      //See if we can produce an english description of the error
      CString sErrorCode(sResponse.Mid(nExclaimationPos+1, 2));

      //Convert to a numeric value
      int nError = 0;
      _stscanf_s(sErrorCode, _T("%x"), &nError);
      if (!sErrorDescription.LoadString(IDS_REDRAT_BASE_ERRORCODE + nError))
        sErrorDescription.LoadString(IDS_REDRAT_BASE_ERRORCODE);
    }
  }

  return bSuccess;
}

UINT CRedRatConnection::MonitorThread()
{
  HANDLE Handles[2];
  Handles[0] = m_StopEvent;
  Handles[1] = m_PortEvent;
  CString sSerialString;

  BOOL bWantStop = FALSE;
  while (!bWantStop)
  {
    //Wait for some data to become available on the port
    try
    {
      //Lock down the serial port
      CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

      DWORD dwEvents = 0;
      m_Port.WaitEvent(dwEvents, m_Overlapped);
    }
    catch(CSerialException* pEx)
    {
      pEx->Delete();
    }

    //WaitCommEvent failed (because it is to be completed overlapped), so wait for 
    //any of the events we are interested in
    DWORD dwWait = WaitForMultipleObjects(2, Handles, FALSE, INFINITE);
    int nSignaledHandle = dwWait - WAIT_OBJECT_0;
    if (nSignaledHandle == 0) //It was the stop request
      bWantStop = TRUE;
    else if(nSignaledHandle == 1) //It was some data to be read from the port
      HandleDataFromPort(sSerialString);
  }

  return 0;
}

void CRedRatConnection::TerminateMonitoringThread()
{
  CComCritSecLock<CComAutoCriticalSection> lock(m_csThread);

  if (m_pMonitorThread)
  {
    //Request the worker thread to exit
    m_StopEvent.SetEvent();

    //Synchronise with the exit of the monitor thread
    WaitForSingleObject(m_pMonitorThread->m_hThread, INFINITE);
    delete m_pMonitorThread;
    m_pMonitorThread = NULL;
  }
}

HRESULT CRedRatConnection::StartMonitoringThread()
{
  //Send the start capture request
  HRESULT hr = _StartCaptureMode();
  if (FAILED(hr))
    return hr;

  CComCritSecLock<CComAutoCriticalSection> lock(m_csThread);

  //Create the monitoring thread
  ATLASSERT(m_pMonitorThread == NULL);
  m_pMonitorThread = AfxBeginThread(_MonitorThread, this, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);  
  if (m_pMonitorThread == NULL)
  {
    Close();

    //Form the rich error information
    return Error(IDS_FAIL_CREATE_THREAD, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, GetLastError()));
  }

  //We are looking after the shutdown of the thread
  m_pMonitorThread->m_bAutoDelete = FALSE;
  m_StopEvent.ResetEvent();
  m_pMonitorThread->ResumeThread();

  return S_OK;
}

void CRedRatConnection::HandleDataFromPort(CString& sSerialString)
{
  //Lock down the serial port
  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  //Read whatever data is present from the serial port
  DWORD dwBytesRead = 0;
  char szBuffer[100];
  try
  {
    if (!m_Port.Read(szBuffer, sizeof(szBuffer) - 1, m_Overlapped, &dwBytesRead))
      m_Port.GetOverlappedResult(m_Overlapped, dwBytesRead, TRUE);
  }
  catch(CSerialException* pEx)
  {
    pEx->Delete();
  }

  //If any data was read from the serial port, then parse the received data
  if (dwBytesRead > 0) 
  {
    //Null terminate the received data
    szBuffer[dwBytesRead] = '\0';

    //Append the newly read data to the buffer
    sSerialString += szBuffer;

    //See if we've got a sentence yet
    int nLocationResponseTerminator = sSerialString.Find(_T("\r\n>"));
    while (nLocationResponseTerminator != -1) 
    {
      //Work out the current signal
      CString sSignal = sSerialString.Left(nLocationResponseTerminator);
      int nSignalLength = sSignal.GetLength();
      if (nSignalLength)
      {
        //Remove the start message indicator
        int nMessageStart = sSignal.Find(_T('['));
        if (nMessageStart == 0)
        {
          nSignalLength--;
          sSignal = sSignal.Right(nSignalLength);
        }

        //Remove the end message indicator
        int nMessageEnd = sSignal.ReverseFind(_T(']'));
        if (nMessageEnd != -1)
        {
          nSignalLength = nMessageEnd;
          sSignal = sSignal.Left(nMessageEnd);
        }

        //Remove the optional checksum
        int nChecksum = sSignal.ReverseFind(_T('#'));
        if (nChecksum != -1)
          sSignal = sSignal.Left(nChecksum);

        //Inform all connected clients of the signal
        BSTR bstrSignal = sSignal.AllocSysString();
        if (bstrSignal)
        {
          Fire_OnSignal(bstrSignal);
          SysFreeString(bstrSignal);
        }
      }

      //Now trim off the signal from what came in from the serial port, and loop back looking
      //for more signals
      sSerialString = sSerialString.Right(sSerialString.GetLength() - (nLocationResponseTerminator + 3));
      nLocationResponseTerminator = sSerialString.Find(_T("\r\n>"));
    }
  }
}

STDMETHODIMP CRedRatConnection::get_Port(long* pVal)
{
  //validate our parameters
	if (pVal == NULL)
    return E_POINTER;

  *pVal = m_nPort;

	return S_OK;
}

STDMETHODIMP CRedRatConnection::get_Timeout(long* pVal)
{
  //Validate our parameters
  if (pVal == NULL)
    return E_POINTER;

  InterlockedExchange(pVal, m_lTimeout);

  return S_OK;
}

STDMETHODIMP CRedRatConnection::put_Timeout(long Val)
{
  InterlockedExchange(&m_lTimeout, Val);

  return S_OK;
}

STDMETHODIMP CRedRatConnection::get_AutoReceiveSignals(VARIANT_BOOL* pVal)
{
  //Validate our parameters
  if (pVal == NULL)
    return E_POINTER;

  LONG volatile bAutoReceiveSignals;
  InterlockedExchange(&bAutoReceiveSignals, m_bAutoReceiveSignals);
  *pVal = bAutoReceiveSignals ? VARIANT_TRUE : VARIANT_FALSE;

  return S_OK;
}

STDMETHODIMP CRedRatConnection::put_AutoReceiveSignals(VARIANT_BOOL Val)
{
  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  HRESULT hr = S_OK;
  if (Val)
  {
    hr = WriteToEEPROM(AUTO_RECEIVE_EEPROM_LOCATION, 1);
    if (FAILED(hr))
      return hr;

    InterlockedExchange(&m_bAutoReceiveSignals, TRUE);
    hr = StartMonitoringThread();
  }
  else
  {
    InterlockedExchange(&m_bAutoReceiveSignals, FALSE);
    TerminateMonitoringThread();
  }

  return hr;
}

STDMETHODIMP CRedRatConnection::get_FirmwareDetails(BSTR* pVal)
{
  TerminateMonitoringThread();

  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  //Send the request
  try
  {
    m_Port.Write("[I]", 3, m_Overlapped);
    DWORD dwBytesWritten = 0;
    m_Port.GetOverlappedResult(m_Overlapped, dwBytesWritten, TRUE);
  }
  catch(CSerialException* pEx)
  {
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    //Form the rich error information
    return Error(IDS_FAIL_WRITE_FIRMWAREDETAILS, IID_IRedRatConnection, hr);
  }

  //Receive the response
  CString sResponse;
  CString sErrorDescription;
  BOOL bError;
  if (!ReadResponse(sResponse, bError, sErrorDescription))
  {
    //Form the rich error information
    return Error(IDS_FAIL_READ_FIRMWAREDETAILS1, IID_IRedRatConnection);
  }

  if (bError)
  {
    //Form the rich error information
    CString sError;
    AfxFormatString2(sError, IDS_FAIL_READ_FIRMWAREDETAILS2, sResponse, sErrorDescription);
    return Error(sError, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_FAIL_READ_FIRMWAREDETAILS2));
  }

  try
  {
    *pVal = sResponse.AllocSysString();
  }
  catch(CMemoryException* pEx)
  {
    pEx->Delete();

    return E_OUTOFMEMORY;
  }

  //Start background monitoring for incoming signals if configured to do so
  if (m_bAutoReceiveSignals)
  {
    HRESULT hr = StartMonitoringThread();
    if (FAILED(hr))
      return hr;
  }

  return S_OK;
}

STDMETHODIMP CRedRatConnection::Send(BSTR Signal)
{
  TerminateMonitoringThread();

  //Form the request string
  CStringA sRequest;
  sRequest.Format("[P%s]", CW2A(Signal));

  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  //Send the request
  try
  {
    m_Port.Write(sRequest.operator LPCSTR(), sRequest.GetLength(), m_Overlapped);
    DWORD dwBytesWritten = 0;
    m_Port.GetOverlappedResult(m_Overlapped, dwBytesWritten, TRUE);
  }
  catch(CSerialException* pEx)
  {
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    //Form the rich error information
    return Error(IDS_FAIL_WRITE_SENDREQUEST, IID_IRedRatConnection, hr);
  }

  //Receive the response
  CString sResponse;
  CString sErrorDescription;
  BOOL bError;
  if (!ReadResponse(sResponse, bError, sErrorDescription))
  {
    //Form the rich error information
    return Error(IDS_FAIL_READ_SENDRESPONSE1, IID_IRedRatConnection);
  }

  if (bError)
  {
    //Form the rich error information
    CString sError;
    AfxFormatString2(sError, IDS_FAIL_READ_SENDRESPONSE2, sResponse, sErrorDescription);
    return Error(sError, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_FAIL_READ_SENDRESPONSE2));
  }

  //Start background monitoring for incoming signals if configured to do so
  if (m_bAutoReceiveSignals)
  {
    HRESULT hr = StartMonitoringThread();
    if (FAILED(hr))
      return hr;
  }

  return S_OK;
}

HRESULT CRedRatConnection::_StartCaptureMode()
{
  //Send the request
  try
  {
    m_Port.Write("[S]", 3, m_Overlapped);
    DWORD dwBytesWritten = 0;
    m_Port.GetOverlappedResult(m_Overlapped, dwBytesWritten, TRUE);
  }
  catch(CSerialException* pEx)
  {
    //Form the rich error information
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    return Error(IDS_FAIL_WRITE_STARTCAPTUREMODE, IID_IRedRatConnection, hr);
  }

  //Receive the response
  CString sResponse;
  CString sErrorDescription;
  BOOL bError;
  if (!ReadResponse(sResponse, bError, sErrorDescription))
  {
    //Form the rich error information
    return Error(IDS_FAIL_READ_STARTCAPTUREMODE1, IID_IRedRatConnection);
  }

  if (bError)
  {
    //Form the rich error information
    CString sError;
    AfxFormatString2(sError, IDS_FAIL_READ_STARTCAPTUREMODE2, sResponse, sErrorDescription);
    return Error(sError, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_FAIL_READ_STARTCAPTUREMODE2));
  }

  return S_OK;
}

STDMETHODIMP CRedRatConnection::StartCaptureMode()
{
  TerminateMonitoringThread();

  //Start background monitoring for incoming signals
  return StartMonitoringThread();
}

STDMETHODIMP CRedRatConnection::TerminateCaptureMode()
{
  TerminateMonitoringThread();

  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  //Send the request
  try
  {
    m_Port.Write("[T]", 3, m_Overlapped);
    DWORD dwBytesWritten = 0;
    m_Port.GetOverlappedResult(m_Overlapped, dwBytesWritten, TRUE);
  }
  catch(CSerialException* pEx)
  {
    //Form the rich error information
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    return Error(IDS_FAIL_WRITE_TERMINATECAPTUREMODE, IID_IRedRatConnection, hr);
  }

  //Receive the response
  CString sResponse;
  CString sErrorDescription;
  BOOL bError;
  if (!ReadResponse(sResponse, bError, sErrorDescription))
  {
    //Form the rich error information
    return Error(IDS_FAIL_READ_TERMINATECAPTUREMODE1, IID_IRedRatConnection);
  }

  if (bError)
  {
    //Form the rich error information
    CString sError;
    AfxFormatString2(sError, IDS_FAIL_READ_TERMINATECAPTUREMODE2, sResponse, sErrorDescription);
    return Error(sError, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_FAIL_READ_TERMINATECAPTUREMODE2));
  }

  return S_OK;
}

STDMETHODIMP CRedRatConnection::Blink()
{
  TerminateMonitoringThread();

  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  //Send the request
  try
  {
    m_Port.Write("[B]", 3, m_Overlapped);
    DWORD dwBytesWritten = 0;
    m_Port.GetOverlappedResult(m_Overlapped, dwBytesWritten, TRUE);
  }
  catch(CSerialException* pEx)
  {
    //Form the rich error information
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    return Error(IDS_FAIL_WRITE_BLINK, IID_IRedRatConnection, hr);
  }

  //Receive the response
  CString sResponse;
  CString sErrorDescription;
  BOOL bError;
  if (!ReadResponse(sResponse, bError, sErrorDescription))
  {
    //Form the rich error information
    return Error(IDS_FAIL_READ_BLINK1, IID_IRedRatConnection);
  }

  if (bError)
  {
    //Form the rich error information
    CString sError;
    AfxFormatString2(sError, IDS_FAIL_READ_BLINK2, sResponse, sErrorDescription);
    return Error(sError, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_FAIL_READ_BLINK2));
  }

  //Start background monitoring for incoming signals if configured to do so
  if (m_bAutoReceiveSignals)
  {
    HRESULT hr = StartMonitoringThread();
    if (FAILED(hr))
      return hr;
  }

  return S_OK;
}

STDMETHODIMP CRedRatConnection::ReadFromEEPROM(long Location, long* pVal)
{
  //Validate our parameters
  if (pVal == NULL)
    return E_POINTER;

  TerminateMonitoringThread();

  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  CStringA sRequest;
  sRequest.Format("[ER%02X]", Location);

  //Send the request
  try
  {
    m_Port.Write(sRequest.operator LPCSTR(), sRequest.GetLength(), m_Overlapped);
    DWORD dwBytesWritten = 0;
    m_Port.GetOverlappedResult(m_Overlapped, dwBytesWritten, TRUE);
  }
  catch(CSerialException* pEx)
  {
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    //Form the rich error information
    return Error(IDS_FAIL_WRITE_READEEPROMREQUEST, IID_IRedRatConnection, hr);
  }

  //Receive the response
  CString sResponse;
  CString sErrorDescription;
  BOOL bError;
  if (!ReadResponse(sResponse, bError, sErrorDescription))
  {
    //Form the rich error information
    return Error(IDS_FAIL_READ_READEEPROMRESPONSE1, IID_IRedRatConnection);
  }

  if (bError)
  {
    //Form the rich error information
    CString sError;
    AfxFormatString2(sError, IDS_FAIL_READ_READEEPROMRESPONSE2, sResponse, sErrorDescription);
    return Error(sError, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_FAIL_READ_READEEPROMRESPONSE2));
  }

  //Parse the response
  if (sResponse.Find(_T('E')) == 0)
    sResponse = sResponse.Right(sResponse.GetLength() - 1);
  _stscanf_s(sResponse, _T("%02X"), pVal);

  //Start background monitoring for incoming signals if configured to do so
  if (m_bAutoReceiveSignals)
  {
    HRESULT hr = StartMonitoringThread();
    if (FAILED(hr))
      return hr;
  }

  return S_OK;
}

STDMETHODIMP CRedRatConnection::WriteToEEPROM(long Location, long Value)
{
  TerminateMonitoringThread();

  CComCritSecLock<CComAutoCriticalSection> lock(m_csPort);

  CStringA sRequest;
  sRequest.Format("[EW%02X%02X]", Location, Value);

  //Send the request
  try
  {
    m_Port.Write(sRequest.operator LPCSTR(), sRequest.GetLength(), m_Overlapped);
    DWORD dwBytesWritten = 0;
    m_Port.GetOverlappedResult(m_Overlapped, dwBytesWritten, TRUE);
  }
  catch(CSerialException* pEx)
  {
    HRESULT hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, pEx->m_dwError);
    pEx->Delete();

    //Form the rich error information
    return Error(IDS_FAIL_WRITE_WRITEEEPROMREQUEST, IID_IRedRatConnection, hr);
  }

  //Receive the response
  CString sResponse;
  CString sErrorDescription;
  BOOL bError;
  if (!ReadResponse(sResponse, bError, sErrorDescription))
  {
    //Form the rich error information
    return Error(IDS_FAIL_READ_WRITEEEPROMRESPONSE1, IID_IRedRatConnection);
  }

  if (bError)
  {
    //Form the rich error information
    CString sError;
    AfxFormatString2(sError, IDS_FAIL_READ_WRITEEEPROMRESPONSE2, sResponse, sErrorDescription);
    return Error(sError, IID_IRedRatConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_FAIL_READ_WRITEEEPROMRESPONSE2));
  }

  //Start background monitoring for incoming signals if configured to do so
  if (m_bAutoReceiveSignals)
  {
    HRESULT hr = StartMonitoringThread();
    if (FAILED(hr))
      return hr;
  }

  return S_OK;
}

